"""Created on Sat Apr 24 20:57:45 2021.

@author: mofarrag
"""
# import Hapi.rrm.calibration as calibration
# import Hapi.rrm.distparameters as distparameters
# import Hapi.rrm.distrrm as distrrm
# import Hapi.rrm.hbv as hbv
# import Hapi.rrm.hbv_bergestrom92 as hbv_bergestrom92
# import Hapi.rrm.hbv_lake as hbv_lake
# import Hapi.rrm.hbvlumped as hbvlumped
# import Hapi.rrm.hbvold as hbvold
# import Hapi.rrm.inputs as inputs
# import Hapi.rrm.routing as routing
# import Hapi.rrm.wrapper as wrapper

if __name__ == "__main__":
    print("rrm")



"""Created on Wed Mar  3 12:40:23 2021.

@author: mofarrag
"""
import datetime as dt
from typing import Any, Union

import pandas as pd
from loguru import logger
from pandas import DataFrame

from Hapi.hm.river import River


class Interface(River):
    """Interface.

    Interface between the Rainfall runoff model and the Hydraulic model

    Methods
    -------
        1- ReadLateralsTable
        2- ReadLaterals
        3- ReadBoundaryConditionsTable
        4- ReadBoundaryConditions
        5- ListAttributes
    """

    Laterals: DataFrame

    def __init__(
        self,
        name: Any,
        version: int = 3,
        start: str = "1952-1-1",
        days: int = 36890,
        fmt: str = "%Y-%m-%d",
    ):
        if not isinstance(start, str):
            raise ValueError("start argument has to be string")
        if not isinstance(version, int):
            raise ValueError("version argument has to be integer number")
        if not isinstance(days, int):
            raise ValueError("number of days has to be integer number")
        if not isinstance(fmt, str):
            raise ValueError("date format 'fmt' has to be a string")

        self.name = name
        self.version = version
        self.start = dt.datetime.strptime(start, fmt)
        self.end = self.start + dt.timedelta(days=days)
        Ref_ind = pd.date_range(self.start, self.end, freq="D")
        self.ReferenceIndex = pd.DataFrame(index=list(range(1, days + 1)))
        self.ReferenceIndex["date"] = Ref_ind[:-1]

        self.LateralsTable = None
        self.RRMProgression = None
        self.BCTable = None
        self.BC = None
        pass

    def readLateralsTable(
        self, path: str, prefix: str = "lf_xsid", suffix: str = ".txt"
    ) -> None:
        """ReadLateralsTable.

        ReadLateralsTable method reads the laterals file
            laterals file : file contains the xsid of the cross-sections that
            has laterals
        if the user has already read te cross section file, the methos is going
        to add column to the crosssection dataframe attribute and is going to add
        a value of 1 to the xs that has laterals

        Parameters
        ----------
        path : [String], optional
            Path to read the results from.
        suffix: [str]
            any
        prefix: []


        Returns
        -------
        IF: [dataframe attribute]
            dataframe with two columns ["filename", "sxid"]
        """
        try:
            self.LateralsTable = pd.read_csv(path, skiprows=[0], header=None)
        except pd.errors.EmptyDataError:
            self.LateralsTable = pd.DataFrame()
            logger.warning("The Lateral table file is empty")
            return

        self.LateralsTable.columns = ["filename"]
        l1 = len(prefix)
        l2 = len(suffix)
        self.LateralsTable["xsid"] = [
            int(i[l1 : len(i) - l2])
            for i in self.LateralsTable[self.LateralsTable.columns[0]]
        ]

        if hasattr(self, "crosssections"):
            self.crosssections["lateral"] = 0
            for i in range(len(self.crosssections)):
                if (
                    self.crosssections.loc[i, "xsid"]
                    in self.LateralsTable["xsid"].tolist()
                ):
                    self.crosssections.loc[i, "lateral"] = 1
        else:
            raise ValueError(
                "Please read the cross section file first using the method 'ReadCrossSections'"
            )

    def readLaterals(
        self,
        fromday: Union[str, int] = "",
        today: Union[str, int] = "",
        path: str = "",
        date_format: str = "'%Y-%m-%d'",
    ):
        """readLaterals.

            read the upstream hydrograph

        Parameters
        ----------
        fromday : [integer], optional
                the day you want to read the result from, the first day is 1 not zero.The default is ''.
        today : [integer], optional
                the day you want to read the result to.
        path : [String], optional
            path to read the results from. The default is ''.
        date_format : "TYPE, optional
            format of the given date string. The default is "'%Y-%m-%d'".

        Returns
        -------
        Laterals : [dataframe attribute].
            dataframe contains the hydrograph of each of the upstream segments
            with segment id as a column name and a column 'total' contains the
            sum of all the hydrographs.
        """
        if not isinstance(self.LateralsTable, DataFrame):
            raise ValueError(
                "Please read the laterals table first using the 'ReadLateralsTable' method"
            )

        self.Laterals = pd.DataFrame()

        if len(self.LateralsTable) > 0:

            for i in range(len(self.LateralsTable)):
                NodeID = self.LateralsTable.loc[i, "xsid"]
                fname = f"lf_xsid{NodeID}"

                self.Laterals[NodeID] = self.readRRMResults(
                    self.version,
                    self.ReferenceIndex,
                    path,
                    fname,
                    fromday,
                    today,
                    date_format,
                )[fname].tolist()
                logger.info(f"Lateral file {fname} is read")

            self.Laterals["total"] = self.Laterals.sum(axis=1)
            if fromday == "":
                fromday = 1
            if today == "":
                today = len(self.Laterals[self.Laterals.columns[0]])

            start = self.ReferenceIndex.loc[fromday, "date"]
            end = self.ReferenceIndex.loc[today, "date"]

            self.Laterals.index = pd.date_range(start, end, freq="D")
        else:
            logger.info("There are no Laterals table please check")

    def readRRMProgression(
        self,
        fromday: int = "",
        today: int = "",
        path: str = "",
        date_format: str = "'%Y-%m-%d'",
    ):
        """ReadRRMProgression.

        read the routed hydrograph by mHM at the location of the lateral
        cross-sections

        Parameters
        ----------
        fromday: [integer], optional
                the day you want to read the result from, the first day is 1 not zero.The default is ''.
        today: [integer], optional
                the day you want to read the result to.
        path: [String], optional
            path to read the results from. The default is ''.
        date_format: "TYPE, optional
            DESCRIPTION. The default is "'%Y-%m-%d'".

        Returns
        -------
        USHydrographs : [dataframe attribute].
            dataframe contains the hydrograph of each of the upstream segments
            with segment id as a column name and a column 'total' contains the
            sum of all the hydrographs.
        """
        if self.LateralsTable is None:
            raise ValueError(
                "Please read the laterals table first using the"
                "'ReadLateralsTable' method "
            )

        self.RRMProgression = pd.DataFrame()

        if len(self.LateralsTable) > 0:

            for i in range(len(self.LateralsTable)):
                NodeID = self.LateralsTable.loc[i, "xsid"]
                fname = "lf_xsid" + str(NodeID)

                self.RRMProgression[NodeID] = self.readRRMResults(
                    self.version,
                    self.ReferenceIndex,
                    path,
                    fname,
                    fromday,
                    today,
                    date_format,
                )[fname].tolist()
                logger.info("RRM Progression file " + fname + " is read")

            # self.RRMProgression['total'] = self.Laterals.sum(axis=1)
            if fromday == "":
                fromday = 1
            if today == "":
                today = len(self.RRMProgression[self.RRMProgression.columns[0]])

            start = self.ReferenceIndex.loc[fromday, "date"]
            end = self.ReferenceIndex.loc[today, "date"]

            self.RRMProgression.index = pd.date_range(start, end, freq="D")

    def readBoundaryConditionsTable(self, path, prefix="bc_xsid", suffix=".txt"):
        """readBoundaryConditionsTable.

        ReadLateralsTable method reads the laterals file
            laterals file : file contains the xsid of the cross-sections that
            has laterals
        if the user has already read te cross section file, the methos is going
        to add column to the crosssection dataframe attribute and is going to add
        a value of 1 to the xs that has laterals

        Parameters
        ----------
        path : [String], optional
            path to read the results from.
        suffix: [string]
            if the lateral files has a suffix in their names
        prefix: [string]
            `if the lateral files has a prefix in their names

        Returns
        -------
        None.
        """
        self.BCTable = pd.read_csv(path, skiprows=[0], header=None)
        self.BCTable.columns = ["filename"]
        l1 = len(prefix)
        l2 = len(suffix)
        self.BCTable["id"] = [
            int(i[l1 : len(i) - l2]) for i in self.BCTable[self.BCTable.columns[0]]
        ]

    def readBoundaryConditions(
        self,
        fromday: Union[str, int] = "",
        today: Union[str, int] = "",
        path: str = "",
        date_format: str = "'%Y-%m-%d'",
    ):
        """ReadUSHydrograph.

        Read the boundary condition

        Parameters
        ----------
        1-fromday : [integer], optional
                the day you want to read the result from, the first day is 1 not zero.The default is ''.
        2-today : [integer], optional
                the day you want to read the result to.
        3-path : [String], optional
            path to read the results from. The default is ''.
        4-date_format : "TYPE, optional
            DESCRIPTION. The default is "'%Y-%m-%d'".

        Returns
        -------
        1-USHydrographs : [dataframe attribute].
            dataframe contains the hydrograph of each of the upstream segments
            with segment id as a column name and a column 'total' contains the
            sum of all the hydrographs.
        """
        assert isinstance(
            self.BCTable, DataFrame
        ), "Please read the lateras table first using the 'ReadLateralsTable' method"

        # if path == '':
        # path = self.CustomizedRunsPath

        self.BC = pd.DataFrame()

        for i in range(len(self.BCTable)):
            NodeID = self.BCTable.loc[i, "id"]
            fname = "bc_xsid" + str(NodeID)
            self.BC[NodeID] = self.readRRMResults(
                self.version,
                self.ReferenceIndex,
                path,
                fname,
                fromday,
                today,
                date_format,
            )[fname].tolist()

            logger.debug("BC file " + fname + " is read")

        self.BC["total"] = self.BC.sum(axis=1)
        if fromday == "":
            fromday = 1
        if today == "":
            today = len(self.BC[NodeID])

        start = self.ReferenceIndex.loc[fromday, "date"]
        end = self.ReferenceIndex.loc[today, "date"]

        self.BC.index = pd.date_range(start, end, freq="D")

    def ListAttributes(self):
        """ListAttributes.

        Print Attributes List
        """
        logger.debug("\n")
        logger.debug(
            f"Attributes List of: {repr(self.__dict__['name'])} - {self.__class__.__name__} Instance\n"
        )
        self_keys = list(self.__dict__.keys())
        self_keys.sort()
        for key in self_keys:
            if key != "name":
                logger.debug(str(key) + " : " + repr(self.__dict__[key]))

        logger.debug("\n")



"""Lumped Conceptual HBV model.

HBV is lumped conceptual model consists of precipitation, snow melt,
soil moisture and response subroutine to convert precipitation into o runoff,
where state variables are updated each time step to represent a specific
hydrologic behaviour the catchment

This version was edited based on a Master Thesis on "Spatio-temporal simulation
of catchment response based on dynamic weighting of hydrological models" on april 2018

- Model inputs are Precipitation, evapotranspiration and temperature, initial
    state variables, and initial discharge.
- Model output is Qalculated dicharge at time t+1
- Model equations are solved using explicit scheme
- model structure uses 15 parameters if the catchment has snow
    [tt, rfcf, sfcf, cfmax, cwh, cfr, fc, beta, e_corr, lp, c_flux, k, k1,
    alpha, perc]

    otherwise it uses 10 parameters
    [rfcf, fc, beta, etf, lp, c_flux, k, k1, alpha, perc]

this HBV is based on Bergstrom, 1992 two reservoirs with three linear responses
surface runoff, interflow and baseflow
"""
# libraries
import numpy as np

DEF_ST = [0.0, 10.0, 10.0, 10.0, 0.0]
DEF_q0 = 0


def Precipitation(prec, temp, tt, rfcf, sfcf):
    """Precipitaiton routine of the HBV96 model. If temperature is lower than TT [degree C], all the precipitation is considered as snow. If the temperature is higher than tt, all the precipitation is considered as rainfall.

    Parameters
    ----------
    temp : float
        Measured temperature [C]
    tt : float
        Lower temperature treshold [C]
    prec : float
        Precipitation [mm]
    rfcf : float
        Rainfall corrector factor
    sfcf : float
        Snowfall corrector factor

    Returns
    -------
    rf : float
        Rainfall [mm]
    sf : float
        Snowfall [mm]
    """

    if temp <= tt:  # if temp <= lower temp threshold
        rf = 0.0  # no rainfall all the precipitation will convert into snowfall
        sf = prec * sfcf

    else:  # temp >= tt: # if temp > upper threshold
        rf = prec * rfcf  # no snowfall all the precipitation becomes rainfall
        sf = 0.0

    return rf, sf


def Snow(temp, rf, sf, wc_old, sp_old, tt, cfmax, cfr, cwh):
    """Snow routine. The snow pack consists of two states: Water Content (wc) and Snow Pack (sp). The first corresponds to the liquid part of the water in the snow, while the latter corresponds to the solid part. If the temperature is higher than the melting point, the snow pack will melt and the solid snow will become liquid. In the opposite case, the liquid part of the snow will refreeze, and turn into solid. The water that cannot be stored by the solid part of the snow pack will drain into the soil as part of infiltration.

        All precipitation simulated to be snow, i.e. falling when the temperature
        is bellow TT, is multiplied by a snowfall correction factor, SFCF [-].

        Snowmelt is calculated with the degree-day method (cfmax). Meltwater and
        rainfall is retained within the snowpack until it exceeds a certain
        fraction, CWH [%] of the water equivalent of the snow

        Liquid water within the snowpack refreezes using cfr

    Parameters
    ----------
    cfmax : float
        Day degree factor
    temp : float
        Temperature [C]
    tt : float
        Temperature treshold for Melting [C]
    cfr : float
        Refreezing factor
    cwh : float
        Capacity for water holding in snow pack [%]
    rf : float
        Rainfall [mm]
    sf : float
        Snowfall [mm]
    wc_old : float
        Water content in previous state [mm]
    sp_old : float
        snow pack in previous state [mm]

    Returns
    -------
    in : float
        Infiltration [mm]
    wc_new : float
        Liquid Water content in the snow[mm]
    sp_new : float
        Snowpack in posterior state [mm]
    """

    if temp > tt:  # if temp > melting threshold
        # then either some snow will melt or the entire snow will melt
        if (
            cfmax * (temp - tt) < sp_old + sf
        ):  # if amount of melted snow < the entire existing snow (previous amount+new)
            melt = cfmax * (temp - tt)
        else:  # if amount of melted snow > the entire existing snow (previous amount+new)
            melt = (
                sp_old + sf
            )  # then the entire existing snow will melt (old snow pack + the current snowfall)

        sp_new = sp_old + sf - melt
        wc_int = wc_old + melt + rf

    else:  # if temp < melting threshold
        # then either some water will freeze or all the water willfreeze
        if (
            cfr * cfmax * (tt - temp) < wc_old + rf
        ):  # then either some water will freeze or all the water willfreeze
            refr = (
                cfr * cfmax * (tt - temp)
            )  # cfmax*(ttm-temp) is the rate of melting of snow while cfr*cfmax*(ttm-temp) is the rate of freeze of melted water  (rate of freezing > rate of melting)
        else:  # if the amount of frozen water > entire water available
            refr = wc_old + rf

        sp_new = sp_old + sf + refr
        wc_int = wc_old - refr + rf

    #
    if wc_int > cwh * sp_new:  # if water content > holding water capacity of the snow
        inf = wc_int - cwh * sp_new  # water content  will infiltrate
        wc_new = cwh * sp_new  # and the capacity of snow of holding water will retained
    else:  # if water content < holding water capacity of the snow
        inf = 0.0  # no infiltration
        wc_new = wc_int

    return inf, wc_new, sp_new


def Soil(temp, inf, ep, sm_old, uz_old, tm, fc, beta, e_corr, lp):
    """Soil soil routine of the HBV-96 model. The model checks for the amount of water that can infiltrate the soil, coming from the liquid precipitation and the snow pack melting. A part of the water will be stored as soil moisture, while other will become runoff, and routed to the upper zone tank.

        -Actual evaporation from the soil box equals the potential evaporation if
        SM/FC is above LP [%] while a linear reduction is used when SM/FC is below
        LP.
        -Groundwater recharge (r) is added to the upper groundwater box (UZ [mm])
        -PERC [mm d-1]defines the maximum percolation rate from the upper to the
        lower groundwater box (LZ [mm])

    Parameters
    ----------
    fc : float
        Filed capacity
    beta : float
        Shape coefficient for effective precipitation separation
    etf : float
        Total potential evapotranspiration
    temp : float
        Temperature
    tm : float
        Average long term temperature
    e_corr : float
        Evapotranspiration corrector factor
    lp : float _soil
        wilting point
    _in : float
        actual infiltration
    ep : float
        actual evapotranspiration
    sm_old : float
        Previous soil moisture value
    uz_old : float
        Previous Upper zone value

    Returns
    -------
    sm_new : float
        New value of soil moisture
    uz_int_1 : float
        New value of direct runoff into upper zone
    """
    # recharge to the upper zone
    r = ((sm_old / fc) ** beta) * inf

    # Adjusted potential evapotranspiration
    ep_int = (1.0 + (temp - tm) * e_corr) * ep

    ea = min(ep_int, (sm_old / (lp * fc)) * ep_int)

    """
    capilary flux related calculations
    # capilary rise
    # cf = c_flux*((fc - sm_old)/fc)

    # if capilary rise is more than what is available take all the available and leave it empty

    # if uz_old + r < cf:
        # cf= uz_old + r
        # uz_int_1=0
    # else:
        # uz_int_1 = uz_old + r - cf

    # sm_new = max(sm_old + inf - r + cf - ea, 0)
    """

    uz_int_1 = uz_old + r
    sm_new = max(sm_old + inf - r - ea, 0)

    return sm_new, uz_int_1


def Response(lz_old, uz_int_1, perc, k, k1, k2, uzl):
    """Response.

        The response routine of the HBV-96 model.
        The response routine is in charge of transforming the current values of
        upper and lower zone into discharge. This routine also controls the
        recharge of the lower zone tank (baseflow). The transformation of units
        also occurs in this point.
        - PERC [mm d-1]defines the maximum percolation rate from the upper to the
        lower groundwater box (SLZ [mm])
        - Runoff from the groundwater boxes is computed as the sum of two or three
        linear outflow equatins depending on whether UZ is above a threshold value,
        UZL [mm], or not

    Parameters
    ----------
    perc : float
        Percolation value [mm\hr]
    k : float
        direct runoff recession coefficient
    k1 : float
        Upper zone recession coefficient
    k2 : float
        Lower zone recession coefficient
    uzl: float
        Upper zone threshold value
    lz_old : float
        Previous lower zone value [mm]
    uz_int_1 : float
        Previous upper zone value before percolation [mm]

    Returns
    -------
    q_uz:[float]
        upper zone discharge (mm/timestep) for both surface runoff & interflow
    q_2:[float]
        Lower zone discharge (mm/timestep).
    """
    # upper zone
    # if perc > Quz then perc = Quz and Quz = 0 if not perc = value and Quz= Quz-perc so take the min
    uz_int_2 = np.max([uz_int_1 - perc, 0.0])

    # surface runoff
    q_0 = k * np.max([uz_int_2 - uzl, 0])

    # Interflow
    q_1 = k1 * uz_int_2

    # as K & k1 are a very small values (0.005) this condition will never happen
    if q_0 + q_1 > uz_int_2:  # if q_0 =30 and UZ=20
        q_0 = uz_int_2 * 0.67  # q_0=20
        q_1 = uz_int_2 * 0.33

    uz_new = uz_int_2 - (q_0 + q_1)

    # lower zone tank
    # if the percolation > upper zone Q all the Quz will percolate
    lz_int_1 = lz_old + np.min([perc, uz_int_1])

    q_2 = k2 * lz_int_1

    if q_2 > lz_int_1:
        q_2 = lz_int_1

    lz_new = lz_int_1 - q_2

    q_uz = q_0 + q_1

    return q_uz, q_2, uz_new, lz_new


def Tf(maxbas):
    """Transfer function weight generator."""
    wi = []
    for x in range(1, maxbas + 1):
        if x <= maxbas / 2.0:
            # Growing transfer
            wi.append(x / (maxbas + 2.0))
        else:
            # Receding transfer
            wi.append(1.0 - (x + 1) / (maxbas + 2.0))

    # Normalise weights
    wi = np.array(wi) / np.sum(wi)
    return wi


def Routing(q, maxbas=1):
    """Routing. This function implements the transfer function using a triangular function.

    Parameters
    ----------
    q: [array]
        discharge array
    maxbas: [int]
        maxbas parameter value.
    """
    assert maxbas >= 1, "Maxbas value has to be larger than 1"
    # Get integer part of maxbas
    maxbas = int(round(maxbas, 0))

    # get the weights
    w = Tf(maxbas)

    # rout the discharge signal
    q_r = np.zeros_like(q, dtype="float64")
    q_temp = q
    for w_i in w:
        q_r += q_temp * w_i
        q_temp = np.insert(q_temp, 0, 0.0)[:-1]

    return q_r


# def StepRun(p, v, St, snow=0):
#     """
#     ============================================================
#         StepRun(p, p2, v, St, snow=0)
#     ============================================================

#     Makes the calculation of next step of discharge and states

#     Parameters
#     ----------
#     p : array_like [18]
#         Parameter vector, set up as:
#         [tt, rfcf, sfcf, cfmax, cfr, cwh, fc, ecorr, etf, lp, k, k1,
#         alpha, beta, cwh, cfr, c_flux, perc, ]
#     p2 : array_like [2]
#         Problem parameter vector setup as:
#         [tfac, area]
#     v : array_like [4]
#         Input vector setup as:
#         [prec, temp, evap, llt]
#     St : array_like [5]
#         Previous model states setup as:
#         [sp, sm, uz, lz, wc]

#     Returns
#     -------
#     q_new : float
#         Discharge [m3/s]
#     St : array_like [5]
#         Posterior model states
#     """
#     ## Parse of parameters from input vector to model
#     if snow==1:
#         # assert len(p) == 16, "current version of HBV (with snow) takes 18 parameter you have entered "+str(len(p))
#         tt = p[0]
#         rfcf = p[1]
#         sfcf = p[2]
#         # snow function
#         cfmax = p[3]
#         cwh = p[4]
#         cfr = p[5]
#         #soil function
#         fc = p[6]
#         beta = p[7]
#         e_corr = p[8]
#         lp = p[9]
#         # response function
#         k = p[10]
#         k1 = p[11]
#         k2 = p[12]
#         uzl = p[13]
#         perc = p[14]

#     elif snow == 0:
#         # assert len(p) >= 11, "current version of HBV (without snow) takes 11 parameter you have entered "+str(len(p))
#         tt = 2.0     # very low but it does not matter as temp is 25 so it is greater than 2
#         rfcf = p[0]    # 1.0 #p[16] # all precipitation becomes rainfall
#         sfcf = 0.00001  # there is no snow
#         # snow function
#         cfmax = 0.00001  # as there is no melting  and sp+sf=zero all the time so it doesn't matter the value of cfmax
#         cwh = 0.00001    # as sp is always zero it doesn't matter all wc will go as inf
#         cfr = 0.000001   # as temp > ttm all the time so it doesn't matter the value of cfr but put it zero
#         #soil function
#         fc = p[1]
#         beta = p[2]
#         e_corr = p[3]
#         lp = p[4]
#         # response function
#         k = p[5]
#         k1 = p[6]
#         k2 = p[7]
#         uzl = p[8]
#         perc = p[9]


#     ## Parse of Inputs
#     prec = v[0] # Precipitation [mm]
#     temp = v[1] # Temperature [C]
#     ep = v[2] # Long terms (monthly) Evapotranspiration [mm]
#     tm = v[3] #Long term (monthly) average temperature [C]

#     ## Parse of states
#     sp_old = St[0]
#     sm_old = St[1]
#     uz_old = St[2]
#     lz_old = St[3]
#     wc_old = St[4]

#     rf, sf = Precipitation(prec, temp, tt, rfcf, sfcf)

#     inf, wc_new, sp_new = Snow(temp, rf, sf, wc_old, sp_old,
#                                tt, cfmax, cfr, cwh)

#     sm_new, uz_int_1 = Soil(temp, inf, ep, sm_old, uz_old, tm,
#                             fc, beta, e_corr, lp)

#     q_uz, q_lz, uz_new, lz_new = Response(lz_old, uz_int_1,
#                                           perc, k, k1, k2, uzl)

#     return q_uz, q_lz, [sp_new, sm_new, uz_new, lz_new, wc_new]


def Simulate(prec, temp, et, ll_temp, par, init_st=None, q_init=None, snow=0):
    """Simulate.

        Run the HBV model for the number of steps (n) in precipitation. The
        resluts are (n+1) simulation of discharge as the model calculates step n+1


    Parameters
    ----------
    prec : array_like [n]
        Average precipitation [mm/h]
    temp : array_like [n]
        Average temperature [C]
    et : array_like [n]
        Potential Evapotranspiration [mm/h]
    par : array_like [18]
        Parameter vector, set up as:
        [ltt, utt, ttm, cfmax, fc, ecorr, etf, lp, k, k1,
        alpha, beta, cwh, cfr, c_flux, perc, rfcf, sfcf]
    init_st : array_like [5], optional
        Initial model states, [sp, sm, uz, lz, wc]. If unspecified,
        [0.0, 30.0, 30.0, 30.0, 0.0] mm
    ll_temp : array_like [n], optional
        Long term average temptearature. If unspecified, calculated from temp.
    q_init : float, optional
        Initial discharge value. If unspecified set to 10.0
    snow: [int]
        1 if snow is considered 0 if not.

    Returns
    -------
    q_sim : array_like [n]
        Discharge for the n time steps of the precipitation vector [m3/s]
    st : array_like [n, 5]
        Model states for the complete time series [mm]
    """
    st = np.zeros([len(prec) + 1, 5], dtype=np.float32)
    q_0 = np.zeros([len(prec) + 1], dtype=np.float32)
    q_1 = np.zeros([len(prec) + 1], dtype=np.float32)
    q_uz = np.zeros([len(prec) + 1], dtype=np.float32)
    q_lz = np.zeros([len(prec) + 1], dtype=np.float32)

    if init_st is None:  # 0  1  2  3  4  5
        st[0, :] = DEF_ST  # [sp,sm,uz,lz,wc,LA]
    else:
        st[0, :] = init_st

    ### initial runoff
    # calculate the runoff for the first time step
    if q_init is None:
        if snow == 1:
            # upper zone
            q_0[0] = par[10] * max(st[0, 2] - par[13], 0)
            q_1[0] = par[11] * st[0, 2]
            q_uz[0] = q_0[0] + q_1[0]
            # lower zone
            q_lz[0] = par[12] * st[0, 3]

        else:
            # upper zone
            q_0[0] = par[5] * max(st[0, 2] - par[8], 0)
            q_1[0] = par[6] * st[0, 2]
            q_uz[0] = q_0[0] + q_1[0]
            # lower zone
            q_lz[0] = par[7] * st[0, 3]
    else:  # if initial runoff value is given distribute it evenlt between upper and lower responses
        q_uz[0] = q_init / 2
        q_lz[0] = q_init / 2

    ## Parse of parameters from input vector to model
    if snow == 1:
        # assert len(p) == 16, "current version of HBV (with snow) takes 18 parameter you have entered "+str(len(p))
        tt = par[0]
        rfcf = par[1]
        sfcf = par[2]
        # snow function
        cfmax = par[3]
        cwh = par[4]
        cfr = par[5]
        # soil function
        fc = par[6]
        beta = par[7]
        e_corr = par[8]
        lp = par[9]
        # response function
        k = par[10]
        k1 = par[11]
        k2 = par[12]
        uzl = par[13]
        perc = par[14]

    elif snow == 0:
        # assert len(par) >= 11, "current version of HBV (without snow) takes 11 parameter you have entered "+str(len(par))
        tt = (
            2.0  # very low but it does not matter as temp is 25 so it is greater than 2
        )
        rfcf = par[0]  # 1.0 #par[16] # all precipitation becomes rainfall
        sfcf = 0.00001  # there is no snow
        # snow function
        # cfmax = 0.00001  # as there is no melting  and sp+sf=zero all the time so it doesn't matter the value of cfmax
        # cwh = 0.00001    # as sp is always zero it doesn't matter all wc will go as inf
        # cfr = 0.000001   # as temp > ttm all the time so it doesn't matter the value of cfr but put it zero
        # soil function
        fc = par[1]
        beta = par[2]
        e_corr = par[3]
        lp = par[4]
        # response function
        k = par[5]
        k1 = par[6]
        k2 = par[7]
        uzl = par[8]
        perc = par[9]

    for i in range(1, len(prec)):
        ## Parse of Inputs
        preci = prec[i]  # Precipitation [mm]
        tempi = temp[i]  # Temperature [C]
        epi = et[i]  # Long terms (monthly) Evapotranspiration [mm]
        tmi = ll_temp[i]  # Long term (monthly) average temperature [C]

        ## Parse of states
        sp_old = st[i - 1, 0]
        sm_old = st[i - 1, 1]
        uz_old = st[i - 1, 2]
        lz_old = st[i - 1, 3]
        wc_old = st[i - 1, 4]

        rf, sf = Precipitation(preci, tempi, tt, rfcf, sfcf)

        if snow == 0:
            inf = rf
            wc_new = 0
            sp_new = 0
        else:
            inf, wc_new, sp_new = Snow(
                tempi, rf, sf, wc_old, sp_old, tt, cfmax, cfr, cwh
            )

        sm_new, uz_int_1 = Soil(
            tempi, inf, epi, sm_old, uz_old, tmi, fc, beta, e_corr, lp
        )

        q_uz[i], q_lz[i], uz_new, lz_new = Response(
            lz_old, uz_int_1, perc, k, k1, k2, uzl
        )

        st[i, :] = [sp_new, sm_new, uz_new, lz_new, wc_new]

        # v = [prec[i], temp[i], et[i], ll_temp[i]]
        # q_uz[i], q_lz[i], st[i,:] = StepRun(par, v, st[i-1,:], snow=snow)

    return q_uz, q_lz, st



"""Routing File contains all Routing functions.

1- Muskingum cunge
2- triangular (MAXBAS) routing function

Created on Sun Apr 29 17:42:04 2018

@author: Mostafa
"""

from typing import Union

import numpy as np


class Routing:
    """Routing class contains routing method.

    Methods
    1- Muskingum
    2- Muskingum_V
    3- TriangularRouting1
        functions :
        1- CalculateWeights
    4- TriangularRouting2
        functions
        1- Tf
    """

    def __init__(self):
        """Routing model does not need any parameters to be instantiated."""
        pass

    @staticmethod
    def Muskingum(inflow, Qinitial, k, x, dt):
        """Muskingum.

        inputs:
        ----------
            1-inflow:
                [numpy array] time series of inflow hydrograph
            2-Qinitial:
                [numeric] initial value for outflow
            3-k:
                [numeric] travelling time (hours)
            4-x:
                [numeric] surface nonlinearity coefficient (0,0.5)
            5-dt:
                [numeric] delta t

        Outputs:
        ----------
            1-outflow:
                [numpy array] time series of routed hydrograph

        examples:
        ----------
        pars[10]=k
        pars[11]=x
        p2[0]=1  # hourly time step
        q_routed = Routing.muskingum(q_uz,q_uz[0],pars[10],pars[11],p2[0])
        """
        c1 = (dt - 2 * k * x) / (2 * k * (1 - x) + dt)
        c2 = (dt + 2 * k * x) / (2 * k * (1 - x) + dt)
        c3 = (2 * k * (1 - x) - dt) / (2 * k * (1 - x) + dt)

        #    if c1+c2+c3!=1:
        #        raise("sim of c1,c2 & c3 is not 1")

        outflow = np.zeros_like(inflow)
        outflow[0] = Qinitial

        for i in range(1, len(inflow)):
            outflow[i] = c1 * inflow[i] + c2 * inflow[i - 1] + c3 * outflow[i - 1]

        outflow = np.round(outflow, 4)

        return outflow

    @staticmethod
    def Muskingum_V(
        inflow: np.ndarray,
        Qinitial: Union[int, float],
        k: Union[int, float],
        x: Union[int, float],
        dt: Union[int, float],
    ) -> np.ndarray:
        """Muskingum_V.

            Vectorized version of Muskingum

        inputs:
        ----------
            1-inflow: [numpy array]
                time series of inflow hydrograph
            2-Qinitial: [numeric]
                initial value for outflow
            3-k: [numeric]
                travelling time (hours)
            4-x: [numeric]
                surface nonlinearity coefficient (0,0.5)
            5-dt: [numeric]
                delta t

        Outputs:
        ----------
            1-outflow:
                [numpy array] time series of routed hydrograph

        examples:
        ----------
        pars[10]=k
        pars[11]=x
        p2[0]=1  # hourly time step
        q_routed = Routing.muskingum(q_uz,q_uz[0],pars[10],pars[11],p2[0])
        """
        c1 = (dt - 2 * k * x) / (2 * k * (1 - x) + dt)
        c2 = (dt + 2 * k * x) / (2 * k * (1 - x) + dt)
        c3 = (2 * k * (1 - x) - dt) / (2 * k * (1 - x) + dt)

        #    if c1+c2+c3!=1:
        #        raise("sim of c1,c2 & c3 is not 1")

        Q = np.zeros_like(inflow)
        Q[0] = Qinitial
        Q[1:] = c1 * np.asarray(inflow[1:]) + c2 * np.asarray(inflow[0:-1])

        for i in range(1, len(inflow)):
            # only if the
            if not Q[i] + c3 * Q[i - 1] < 0:
                Q[i] = Q[i] + c3 * Q[i - 1]

        return Q

    @staticmethod
    def Tf(maxbas):
        """Tf.

            Transfer function weight generator in a shape of a triangle.

        Inputs:
        ----------
            1-maxbas:
                [integer] number of time steps that the triangular routing function
                is going to divide the discharge into, based on the weights
                generated from this function, min value is 1 and default value is 1

        Outputs:
        ----------
            1-wi:
                [numpy array] array of normalised weights

        Example:
        ----------
            ws=Tf(5)
        """

        wi = []
        for x in range(1, maxbas + 1):  # if maxbas=3 so x=[1,2,3]
            if (
                x <= (maxbas) / 2.0
            ):  # x <= 1.5  # half of values will form the rising limb and half falling limb
                # Growing transfer    # rising limb
                wi.append((x) / (maxbas + 2.0))
            else:
                # Receding transfer    # falling limb
                wi.append(1.0 - (x + 1) / (maxbas + 2.0))

        # Normalise weights
        wi = np.array(wi) / np.sum(wi)
        return wi

    @staticmethod
    def TriangularRouting2(q, maxbas=1):
        """TriangularRouting implements the transfer function using a triangular function (considers only integer values of Maxbas parameter)

        Inputs:
        ----------
            1-q:
                [numpy array] time series of discharge hydrographs
            2-maxbas:
                [integer] number of time steps that the triangular routing function
                is going to divide the discharge into, based on the weights
                generated from this function, min value is 1 and default value is 1

        Outputs:
        ----------
            1-q_r:
                [numpy array] time series of routed hydrograph

        examples:
        ----------
            q_sim=TriangularRouting(np.array(q_sim), parameters[-1])
        """
        # input data validation
        assert maxbas >= 1, "Maxbas value has to be larger than 1"

        # Get integer part of maxbas
        maxbas = int(round(maxbas, 0))

        # get the weights
        w = Routing.Tf(maxbas)

        # rout the discharge signal
        q_r = np.zeros_like(q, dtype="float64")
        q_temp = np.float32(q)
        for w_i in w:
            q_r += q_temp * w_i
            q_temp = np.insert(q_temp, 0, 0.0)[:-1]

        return q_r

    @staticmethod
    def CalculateWeights(MAXBAS):
        """CalculateMaxBas calculate the MAXBAS Weights based on a MAXBAX number The MAXBAS is a HBV parameter that controls the routing.

        Inputs:
        ----------
            1-MAXBAS:
                []

        Example:
        ----------
            maxbasW = CalculateMaxBas(5)
            maxbasW =

                0.0800    0.2400    0.3600    0.2400    0.0800

            It is important to mention that this function allows to obtain weights
            not only for interger values but from decimals values as well.
        """
        yant = 0
        Total = 0  # Just to verify how far from the unit is the result

        TotalA = (MAXBAS * MAXBAS * np.sin(np.pi / 3)) / 2
        IntPart = np.floor(MAXBAS)
        RealPart = MAXBAS - IntPart
        PeakPoint = MAXBAS % 2
        flag = 1  # 1 = "up"  ; 2 = down

        if RealPart > 0:  # even number 2,4,6,8,10
            maxbasW = np.ones(int(IntPart) + 1)  # if even add 1
        else:  # odd number
            maxbasW = np.ones(int(IntPart))

        for x in range(int(MAXBAS)):

            if x < (MAXBAS / 2.0) - 1:
                # Integral of  x dx with slope of 60 degree Equilateral triangle
                ynow = np.tan(np.pi / 3) * (x + 1)
                # ' Area / Total Area
                maxbasW[x] = ((ynow + yant) / 2) / TotalA
            else:  # The area here is calculated by the formlua of a trapezoidal (B1+B2)*h /2
                if flag == 1:
                    ynow = np.sin(np.pi / 3) * MAXBAS
                    if PeakPoint == 0:
                        maxbasW[x] = ((ynow + yant) / 2) / TotalA
                    else:
                        A1 = ((ynow + yant) / 2) * (MAXBAS / 2.0 - x) / TotalA
                        yant = ynow
                        ynow = (MAXBAS * np.sin(np.pi / 3)) - (
                            np.tan(np.pi / 3) * (x + 1 - MAXBAS / 2.0)
                        )
                        A2 = ((ynow + yant) * (x + 1 - MAXBAS / 2.0) / 2) / TotalA
                        maxbasW[x] = A1 + A2

                    flag = 2
                else:
                    # 'sum of the two height in the descending part of the triangle
                    ynow = MAXBAS * np.sin(np.pi / 3) - np.tan(np.pi / 3) * (
                        x + 1 - MAXBAS / 2.0
                    )
                    # Multiplying by the height of the trapezoidal and dividing by 2
                    maxbasW[x] = ((ynow + yant) / 2) / TotalA

            Total = Total + maxbasW[x]
            yant = ynow

        x = int(MAXBAS)
        # x = x + 1

        if RealPart > 0:
            if np.floor(MAXBAS) == 0:
                MAXBAS = 1
                maxbasW[x] = 1
                NumberofWeights = 1
            else:
                maxbasW[x] = (yant * (MAXBAS - (x)) / 2) / TotalA
                Total = Total + maxbasW[x]
                NumberofWeights = x
        else:
            NumberofWeights = x - 1

        return maxbasW

    @staticmethod
    def TriangularRouting1(Q, MAXBAS):
        """TriangularRouting1 calculate the routing from a input hydrograph using the MAXBAS parameter from the HBV model (considers float values of Maxbas parameter).

        EXAMPLE:
        ----------
            [Qout,maxbasW]=RoutingMAXBAS(Q,5);
            where:
            Qout = output hydrograph
            maxbasW = MAXBAS weight
            Q = input hydrograph
            5 = MAXBAS parameter value.
        """

        # CALCULATE MAXBAS WEIGHTS
        maxbasW = Routing.CalculateWeights(MAXBAS)

        Qw = np.ones((len(Q), len(maxbasW)))
        # Calculate the matrix discharge
        for i in range(len(Q)):  # 0 to 10
            for k in range(len(maxbasW)):  # 0 to 4
                Qw[i, k] = maxbasW[k] * Q[i]

        def mm(A, s):
            tot = []
            for o in range(np.shape(A)[1]):  # columns
                for t in range(np.shape(A)[0]):  # rows
                    tot.append(A[t, o])
            Su = tot[s:-1:s]
            return Su

        # Calculate routing
        j = 0
        Qout = np.ones(shape=(len(Q)))

        for i in range(len(Q)):
            if i == 0:
                Qout[i] = Qw[i, i]
            elif i < len(maxbasW) - 1:
                A = Qw[0 : i + 1, :]
                s = len(A) - 1  # len(A) is the no of rows or use int(np.shape(A)[0])
                Su = mm(A, s)

                Qout[i] = sum(Su[0 : i + 1])
            else:
                A = Qw[j : i + 1, :]
                s = len(A) - 1
                Su = mm(A, s)
                Qout[i] = sum(Su)
                j = j + 1

        return Qout  # ,maxbasW



"""Created on Tue Feb  4 14:57:30 2020.

@author: mofarrag
"""
import datetime as dt

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pandas.core.frame import DataFrame
from pyramids.raster import Raster

# import os


class Event:
    # class attributes
    """Event. The Event class reads all the results of the Hydraulic model to preform all kind of analysis on flood event basis and the overtopping.

    Methods:
        1- IndexToDate
        2- CreateEventIndex
        3- GetAllEvents
        4- Overtopping
        5- VolumeError
        6- OverlayMaps
        7- ReadEventIndex
        8- Histogram
        9- Drop
        10- Save
        11- GetEventBeginning
        12- GetEventEnd
        13- PrepareForPlotting
        14- ListAttributes
    """

    def __init__(
        self,
        name,
        start="1950-1-1",
        days=36890,
        leftOvertopping_Suffix="_left.txt",
        RightOvertopping_Suffix="_right.txt",
        DepthPrefix="DepthMax",
        DurationPrefix="Duration",
        ReturnPeriodPrefix="ReturnPeriod",
        Compressed=True,
    ):
        """Event. To instantiate the Event class you need to provide the following arguments.

        Parameters
        ----------
        name : [str]
            DESCRIPTION.
        start : [str], optional
            start date. The default is "1950-1-1".
        days : integer, optional
            length of the simulation . The default is 36890.
        leftOvertopping_Suffix : [str], optional
            the prefix you used to name the overtopping form the left bank files.
            The default is "_left.txt".
        RightOvertopping_Suffix : TYPE, optional
            the prefix you used to name the overtopping form the right bank files.
            The default is "_right.txt".
        DepthPrefix : [str], optional
            the prefix you used to name the Max depth raster result maps.
            The default is "DepthMax".
        DurationPrefix : [str], optional
            the prefix you used to name the inundation duration raster result maps.
            The default is "Duration".
        ReturnPeriodPrefix : [str], optional
            the prefix you used to name the Return Period raster result maps.
            The default is "ReturnPeriod".
        Compressed : [bool], optional
            True if the result raster/ascii files are compressed. The default is True.

        Returns
        -------
        None.
        """
        # instance attribute
        self.name = name
        self.start = dt.datetime.strptime(start, "%Y-%m-%d")
        self.end = self.start + dt.timedelta(days=days)

        self.leftOvertopping_Suffix = leftOvertopping_Suffix
        self.RightOvertopping_Suffix = RightOvertopping_Suffix
        self.DepthPrefix = DepthPrefix
        self.DurationPrefix = DurationPrefix
        self.ReturnPeriodPrefix = ReturnPeriodPrefix
        self.TwoDResultPath = ""
        self.Compressed = Compressed
        Ref_ind = pd.date_range(self.start, self.end, freq="D")

        # the last day is not in the results day Ref_ind[-1]
        # write the number of days + 1 as python does not include the last number in the range
        # 19723 days so write 19724
        self.ReferenceIndex = pd.DataFrame(index=list(range(1, days + 1)))
        self.ReferenceIndex["date"] = Ref_ind[:-1]
        # create dictionary to store any extracted values from maps
        self.ExtractedValues = dict()
        self.EventIndex = None
        self.EventBeginning = None
        self.EndDays = None

    # method
    def IndexToDate(self):
        """IndexToDate. get the date coresponding to a given index.

        Returns
        -------
        Date
        """
        # convert the index into date
        dateFn = lambda i: self.ReferenceIndex.loc[i, "date"]
        # get the date the column 'id'
        date = self.EventIndex.loc[:, "id"].to_frame().applymap(dateFn)
        self.EventIndex["date"] = date

    def CreateEventIndex(self, IndexPath: str):
        """CreateEventIndex. CreateEventIndex takes the path to the index file result from the 2D model and creates a data frame to start adding the components of the EventIndex table.

        Parameters
        ---------
        IndexPath: [String]
            path including the file name and extention of the index file result from the 2D model

        Returns
        -------
        EventIndex: [dataframe]
            this method creates an instance attribute of type dataframe with columns ['id','continue', 'IndDiff',
            'Duration']
        """
        # FIXME
        # if the flood event does not have overtopping for 1 day then continues to
        # overtop after the method considers it as two separate events however
        # it is the same event (if the gap is less than 10 days it is still
        # considered the same event)

        # read the index file (containing the id of the days where flood happens (2D
        # algorithm works))
        EventDays = pd.read_csv(IndexPath, header=None)
        EventIndex = EventDays.rename(columns={0: "id"})
        # convert the index into date
        self.EventIndex = EventIndex.loc[:, :]
        self.IndexToDate()

        self.EventIndex.loc[:, "continue"] = 0
        # index difference maybe different than the duration as there might be
        # a gap in the middle of the event
        self.EventIndex.loc[:, "IndDiff"] = 0
        self.EventIndex.loc[:, "Duration"] = 0

        # the first day in the index file is an event beginning
        self.EventBeginning = self.EventIndex.loc[0, "date"]
        for i in range(1, len(self.EventIndex)):
            # if the day is previous day+1
            if self.EventIndex.loc[i, "id"] == self.EventIndex.loc[i - 1, "id"] + 1:
                # then the event continues
                self.EventIndex.loc[i, "continue"] = 1
                # increase the duration
                self.EventIndex.loc[i, "IndDiff"] = (
                    self.EventIndex.loc[i - 1, "IndDiff"] + 1
                )

                self.EventIndex.loc[i, "Duration"] = (
                    self.EventIndex.loc[i, "date"] - self.EventBeginning
                ).days + 1
            else:  # if not then the day is the start of another event
                self.EventBeginning = self.EventIndex.loc[i, "date"]

    def GetAllEvents(self):
        """GetAllEvents. GetAllEvents methods returns the end day of all events.

        Returns
        -------
            None.
        """
        assert hasattr(self, "EventIndex"), "please read/Create the EventIndex"
        IDs = list()
        for i in range(len(self.EventIndex)):
            if self.EventIndex.loc[i, "continue"] == 0 and i != 0:
                IDs.append(self.EventIndex.loc[i - 1, "id"])

        self.EndDays = IDs

    def Overtopping(self, OvertoppingPath: str):
        """Overtopping. Overtopping method reads the overtopping file and check if the EventIndex dataframe has already need created by the CreateEventIndex method, it will add the overtopping to it, if not it will create the EventIndex dataframe.

        Inputs:
            1- OvertoppingPath:
                [String] path including the file name and extention of the Overtopping
                file result from the 1D model
        Outputs:
            1- EventIndex:
                [dataframe] this method creates an instance attribute of type
                dataframe with columns ['id','continue', 'IndDiff', 'Duration',
                'Overtopping', 'OvertoppingCum', 'Volume']
        """
        OverTopTotal = pd.read_csv(OvertoppingPath, delimiter=r"\s+")  # , header = None
        # FIXME
        # if the flood event does not have overtopping for 1 day then continues to
        # overtop after the method considers it as two separate events however
        # it is the same event (if the gap is less than 10 days it is still
        # considered the same event)
        if not isinstance(self.EventIndex, DataFrame):
            # create the dataframe if the user did not use the CreateEventIndex method to
            # create the EventIndex dataframe
            self.EventIndex = pd.DataFrame()
            self.EventIndex["id"] = OverTopTotal["Step"]
            self.IndexToDate()

            self.EventIndex.loc[:, "continue"] = 0
            # index difference maybe different than the duration as there might be
            # a gap in the middle of the event
            self.EventIndex.loc[:, "IndDiff"] = 0
            self.EventIndex.loc[:, "Duration"] = 0

            # the first day in the index file is an event beginning
            self.EventBeginning = self.EventIndex.loc[0, "date"]
            for i in range(1, len(self.EventIndex)):
                # if the day is previous day+1
                if self.EventIndex.loc[i, "id"] == self.EventIndex.loc[i - 1, "id"] + 1:
                    # then the event continues
                    self.EventIndex.loc[i, "continue"] = 1
                    # increase the duration
                    self.EventIndex.loc[i, "IndDiff"] = (
                        self.EventIndex.loc[i - 1, "IndDiff"] + 1
                    )

                    self.EventIndex.loc[i, "Duration"] = (
                        self.EventIndex.loc[i, "date"] - self.EventBeginning
                    ).days + 1
                else:  # if not then the day is the start of another event
                    self.EventBeginning = self.EventIndex.loc[i, "date"]

        # store the overtoppiung data in the EventIndex dataframe
        self.EventIndex["Overtopping"] = OverTopTotal["overtopping(m3/s)"]

        self.EventIndex.loc[0, "OvertoppingCum"] = self.EventIndex.loc[0, "Overtopping"]
        for i in range(1, len(self.EventIndex)):
            if self.EventIndex.loc[i, "continue"] == 0:
                self.EventIndex.loc[i, "OvertoppingCum"] = self.EventIndex.loc[
                    i, "Overtopping"
                ]
            else:
                self.EventIndex.loc[i, "OvertoppingCum"] = (
                    self.EventIndex.loc[i, "Overtopping"]
                    + self.EventIndex.loc[i - 1, "OvertoppingCum"]
                )
        # the volume of water is m3/s for hourly stored and acumulated values
        # volume = overtopping * 60 *60 = m3
        self.EventIndex.loc[:, "Volume"] = (
            self.EventIndex.loc[:, "OvertoppingCum"] * 60 * 60
        )

    def VolumeError(self, Path):
        """VolumeError. VolumeError method reads the VoleError file, assign values to the the coresponding time step.

        Parameters
        ----------
        Path : [String]
            a path to the folder includng the maps.

        Returns
        -------
        EventIndex: [dataframe attribute].
            add columns ['DEMError','StepError','TooMuchWater'] to the EventIndex dataframe
        """
        # read the VolError file
        VolError = pd.read_csv(Path, delimiter=r"\s+")
        self.EventIndex["DEMError"] = 0
        self.EventIndex["StepError"] = 0
        self.EventIndex["TooMuchWater"] = 0

        for i in range(len(VolError)):
            loc = np.where(VolError.loc[i, "step"] == self.EventIndex.loc[:, "id"])[0][
                0
            ]
            self.EventIndex.loc[
                loc, ["DEMError", "StepError", "TooMuchWater"]
            ] = VolError.loc[
                i, ["DEM_Error", "PreviousDepthError", "TOOMuchWaterError"]
            ].tolist()

        self.EventIndex["VolError"] = (
            self.EventIndex["StepError"]
            + self.EventIndex["DEMError"]
            + self.EventIndex["TooMuchWater"]
        )
        self.EventIndex["VolError2"] = self.EventIndex["VolError"] / 20

    def OverlayMaps(self, Path, BaseMapF, ExcludedValue, OccupiedCellsOnly, SavePath):
        """OverlayMaps. OverlayMaps method reads all the maps in the folder given by Path input and overlay them with the basemap and for each value in the basemap it create a dictionary with the intersected values from all maps.

        Parameters
        ----------
        Path: [String]
            a path to the folder includng the maps.
        BaseMapF: [String]
            a path includng the name of the ASCII and extention like
            path="data/cropped.asc"
        ExcludedValue: [Numeric]
            values you want to exclude from exteacted values
        OccupiedCellsOnly: [Bool]
            if you want to count only cells that is not ExcludedValue.
        SavePath: [String]
            a path to the folder to save a text file for each
            value in the base map including all the intersected values
            from other maps.

        Returns
        -------
        ExtractedValues: [Dict]
            dictonary with a list of values in the basemap as keys and for each key a list of all the intersected
            values in the maps from the path
        NonZeroCells: [dataframe]
            dataframe with the first column as the "file" name and the second column is the number of cells in each map
        """
        self.DepthValues, NonZeroCells = Raster.overlayMaps(
            Path,
            BaseMapF,
            self.DepthPrefix,
            ExcludedValue,
            self.Compressed,
            OccupiedCellsOnly,
        )

        # NonZeroCells dataframe with the first column as the "file" name and the second column
        # is the number of cells in each map

        NonZeroCells["days"] = [
            int(i[len(self.DepthPrefix) : -4]) for i in NonZeroCells["files"].tolist()
        ]
        # get the numbe of inundated cells in the Event index data frame
        self.EventIndex["cells"] = 0

        for i in range(len(NonZeroCells)):
            # get the location in the EventIndex dataframe
            try:
                loc = np.where(
                    NonZeroCells.loc[i, "days"] == self.EventIndex.loc[:, "id"]
                )[0][0]
            except IndexError:
                # if it does not find the event in the eventindex table ignore
                continue
            # store number of cells
            self.EventIndex.loc[loc, "cells"] = NonZeroCells.loc[i, "cells"]

        # save depths of each sub-basin
        inundatedSubs = list(self.DepthValues.keys())
        for i in range(len(inundatedSubs)):
            np.savetxt(
                SavePath + "/" + str(inundatedSubs[i]) + ".txt",
                self.DepthValues[inundatedSubs[i]],
                fmt="%4.2f",
            )

    def ReadEventIndex(self, Path):
        """ReadEventIndex ReadEventIndex method reads the EventIndex table created using the "CreateEventIndex" or "Overtopping" methods.

        Parameters
        ----------
        Path : [str]
            Path to the EventIndex file.

        Returns
        -------
        EventIndex : [dataframe].
            dataframe of the EventIndex table
        """
        EventIndex = pd.read_csv(Path)
        self.EventIndex = EventIndex
        self.IndexToDate()

    def Histogram(
        self, Day, ExcludeValue, OccupiedCellsOnly, Map=1, filter1=0.2, filter2=15
    ):
        """Histogram Histogram method extract values fro the event MaxDepth map and plot the histogram th emethod check first if you already extracted the values before then plot the histogram.

        Parameters
        ----------
        Day : [Integer]
            DESCRIPTION.
        ExcludeValue : [Integer]
            DESCRIPTION.
        OccupiedCellsOnly : TYPE
            DESCRIPTION.
        Map : [integer], optional
            1 for the max depth maps, 2 for the duration map, 3 for the
            return period maps. The default is 1.

        Returns
        -------
        None.
        """
        # check if the object has the attribute ExtractedValues
        if hasattr(self, "ExtractedValues"):
            # get the list of event that then object has their Extractedvalues
            if Day not in list(self.ExtractedValues.keys()):
                # depth map
                if Map == 1:
                    Path = self.TwoDResultPath + self.DepthPrefix + str(Day) + ".zip"
                elif Map == 2:
                    Path = self.TwoDResultPath + self.DurationPrefix + str(Day) + ".zip"
                else:
                    Path = (
                        self.TwoDResultPath
                        + self.ReturnPeriodPrefix
                        + str(Day)
                        + ".zip"
                    )

                ExtractedValues, NonZeroCells = Raster.extractValues(
                    Path, ExcludeValue, self.Compressed, OccupiedCellsOnly
                )
                self.ExtractedValues[Day] = ExtractedValues

        ExtractedValues = self.ExtractedValues[Day]
        # filter values
        ExtractedValues = [j for j in ExtractedValues if j > filter1]
        ExtractedValues = [j for j in ExtractedValues if j < filter2]
        # plot
        # fig, ax1 = plt.subplots(figsize=(10,8))
        # ax1.hist(ExtractedValues, bins=15, alpha = 0.4) #width = 0.2,

        n, bins, patches = plt.hist(
            x=ExtractedValues, bins=15, color="#0504aa", alpha=0.7, rwidth=0.85
        )
        plt.grid(axis="y", alpha=0.75)
        plt.xlabel("Value", fontsize=15)
        plt.ylabel("Frequency", fontsize=15)
        plt.xticks(fontsize=15)
        plt.yticks(fontsize=15)

        plt.ylabel("Frequency", fontsize=15)
        plt.tight_layout()
        # plt.title('Normal Distribution Histogram matplotlib',fontsize=15)
        plt.show()
        return n, bins, patches

    def Drop(self, DropList):
        """Drop Drop method deletes columns from the EventIndex dataframe.

        Parameters
        ----------
        DropList: [list]
            list of column names to delete from the EventIndex dataframe table

        Returns
        -------
        EventIndex: [datadrame]
            the EventIndex dataframe without the columns in the Droplist
        """
        dataframe = self.EventIndex.loc[:, :]
        columns = list(dataframe.columns)

        [columns.remove(i) for i in DropList]

        dataframe = dataframe.loc[:, columns]
        self.EventIndex = dataframe

    def Save(self, Path):
        """Save Save method saves the EventIndex table.

        Parameters
        ----------
        Path : [str]
            Path to where you want to save the table.

        Returns
        -------
        None.
        """
        self.EventIndex.to_csv(Path, header=True, index=None)  # index_label = "Index"

    def GetEventBeginning(self, loc):
        """GetEventBeginning. EventBeginning method returns the index of the beginning of the event in the EventIndex dataframe.

        Parameters
        ----------
        loc: [Integer]
            dataframe index of the day you want to trace back to get the begining

        Returns
        -------
        ind: [Integer]
            dataframe index of the beginning day of the event

        Example
        ------
            1- if you want to get the beginning of the event that has the highest
            overtopping
            HighOvertoppingInd = EventIndex['Overtopping'].idxmax()
            ind = EventBeginning(HighOvertoppingInd)
        """
        # loc = np.where(self.EventIndex['id'] == day)[0][0]
        # get all the days in the same event before that day as the inundation in the maps may
        # happen due to any of the days before not in this day
        ind = self.EventIndex.index[loc - self.EventIndex.loc[loc, "IndDiff"]]
        day = self.EventIndex.loc[ind, "id"]
        return ind, day

        # # filter the dataframe and get only the 'indDiff' and 'id' columns
        # FilteredEvent = self.EventIndex.loc[:,['IndDiff','id']]
        # FilteredEvent['diff'] = FilteredEvent.index - ind
        # # get only days before the day you inputed
        # FilteredEvent = FilteredEvent[FilteredEvent['diff'] <=0 ]
        # # start to search from down to up till you get the first 0 in the IndDiff
        # for i in range(self.EventIndex['Duration'].max()):

        #     if FilteredEvent.loc[len(FilteredEvent)-1-i,'IndDiff'] == 0:
        #         break

        # return FilteredEvent.index[len(FilteredEvent)-1-i]

    def GetEventEnd(self, loc):
        """GetEventEnd. method returns the index of the beginning of the event in the EventIndex dataframe.

        Parameters
        ----------
        loc: [Integer]
            dataframe index of the day you want to trace back to get the begining

        Returns
        -------
        ind: [Integer]
            dataframe index of the beginning day of the event

        Example
        -------
        1- if you want to get the beginning of the event that has the highest
        overtopping
        HighOvertoppingInd = EventIndex['Overtopping'].idxmax()
        ind = EventBeginning(HighOvertoppingInd)
        """
        # loc = np.where(self.EventIndex['id'] == day)[0][0]
        # get all the days in the same event before that day as the inundation in the maps may
        # happen due to any of the days before not in this day

        # filter the dataframe and get only the 'indDiff' and 'id' columns
        FilteredEvent = self.EventIndex.loc[:, ["continue", "id"]]
        # get only days before the day you inputed
        for i in range(loc + 1, len(FilteredEvent)):
            # start search from the following day
            if FilteredEvent.loc[i, "continue"] != 1:
                break

        ind = i - 1
        day = self.EventIndex.loc[ind, "id"]

        return ind, day

    def PrepareForPlotting(self, ColumnName):
        """PrepareForPlotting. PrepareForPlotting takes a time series in the EventIndex dataframe and fill the days that does not exist in date column and fill it with zero to properly plot it without letting the graph mislead the viewer of connecting the data over the gap period.

        Parameters
        ----------
        ColumnName : [String]
            name of the columns you want.

        Returns
        -------
        NewDataFrame : [Dataframe]
            dataframe with a date column, and the required column
        """
        NewDataFrame = pd.DataFrame()
        NewDataFrame["date"] = self.ReferenceIndex["date"].tolist()
        NewDataFrame[ColumnName] = 0
        for i in range(len(self.EventIndex)):
            loc = np.where(NewDataFrame["date"] == self.EventIndex.loc[i, "date"])[0][0]
            NewDataFrame.loc[loc, ColumnName] = self.EventIndex.loc[i, ColumnName]

        return NewDataFrame

    def ListAttributes(self):
        """Print Attributes List."""

        print("\n")
        print(
            "Attributes List of: "
            + repr(self.__dict__["name"])
            + " - "
            + self.__class__.__name__
            + " Instance\n"
        )
        self_keys = list(self.__dict__.keys())
        self_keys.sort()
        for key in self_keys:
            if key != "name":
                print(str(key) + " : " + repr(self.__dict__[key]))

        print("\n")


if __name__ == "__main__":
    x = Event("Event")



"""Created on Sun Apr 29 17:17:54 2018.

@author: Mostafa
"""
import numpy as np

from Hapi.rrm import hbv_lake
from Hapi.rrm.distrrm import DistributedRRM as distrrm
from Hapi.rrm.routing import Routing as routing


class Wrapper:
    """Wrapper.

    The class connect different commponent together (lumped run of the distributed with the spatial routing) for Hapi
    and for FW1.

    Methods:
        1- HapiModel
        2- RRMWithlake
        3- FW1
        4- FW1Withlake
        5- Lumped
    """

    def __init__(self):
        """wrapper object does not need any information to be created."""
        pass

    @staticmethod
    def RRMModel(Model, ll_temp=None, q_0=None):
        """RRMModel.

            HapiModel connect two modules :
            1- The distributed rainfall runoff: model runs separately for each cell
            2- The Spatial routing scheme (routing is following river network)

        Inputs:
        ----------
            1-DEM:
                [gdal.dataset] DEM raster file of the catchment (clipped to the catchment only)
            2-flow_acc:
                [gdal.dataset] flow accumulation raster file of the catchment (clipped to the catchment only)
            3-flow_direct:
                [gdal.dataset] flow Direction raster file of the catchment (clipped to the catchment only)
            4-sp_prec:
                [numpy array] 3d array of the precipitation data, sp_prec should
                have the same 2d dimension of raster input
            5-sp_et:
                [numpy array] 3d array of the evapotranspiration data, sp_et should
                have the same 2d dimension of raster input
            6-sp_temp:
                [numpy array] 3d array of the temperature data, sp_temp should
                have the same 2d dimension of raster input
            7-sp_par:
                [numpy array] number of 2d arrays of the catchment properties spatially
                distributed in 2d and the third dimension is the number of parameters,
                sp_pars should have the same 2d dimension of raster input
            8-p2:
                [List] list of unoptimized parameters
                p2[0] = tfac, 1 for hourly, 0.25 for 15 min time step and 24 for daily time step
                p2[1] = catchment area in km2
            9-kub:
                [float] upper bound of K value (traveling time in muskingum routing method)
            10-klb:
                [float] Lower bound of K value (traveling time in muskingum routing method)
            11-init_st:
                [list] initial state variables values [sp, sm, uz, lz, wc]. default=None
            12-ll_temp:
                [numpy array] 3d array of the long term average temperature data
            13-q_0:
                [float] initial discharge m3/s

        Outputs:
        ----------
            1-statevariables:
                [numpy ndarray] 4D array (rows,cols,time,states) states are [sp,wc,sm,uz,lv]
            2-qlz:
                [numpy ndarray] 3D array of the lower zone discharge
            3-quz:
                [numpy ndarray] 3D array of the upper zone discharge
            4-qout:
                [numpy array] 1D timeseries of discharge at the outlet of the catchment
                of unit m3/sec
            5-quz_routed:
                [numpy ndarray] 3D array of the upper zone discharge  accumulated and
                routed at each time step
            6-qlz_translated:
                [numpy ndarray] 3D array of the lower zone discharge translated at each time step
        """
        # run the rainfall runoff model separately
        distrrm.RunLumpedRRM(Model)

        # run the GIS part to rout from cell to another
        distrrm.SpatialRouting(Model)

        # Model.qout = Model.qout[:-1]

    @staticmethod
    def RRMWithlake(Model, Lake, ll_temp=None, q_0=None):
        """RRMWithlake.

            RRMWithlake connects three modules the lake, the distributed
            ranfall-runoff module and spatial routing module

        Parameters
        ----------
        Model : [Catchment object]
            DESCRIPTION.
        Lake : TYPE
            DESCRIPTION.
        ll_temp : TYPE, optional
            DESCRIPTION. The default is None.
        q_0 : TYPE, optional
            DESCRIPTION. The default is None.

        Returns
        -------
        None.
        """

        plake = Lake.MeteoData[:, 0]
        et = Lake.MeteoData[:, 1]
        t = Lake.MeteoData[:, 2]
        tm = Lake.MeteoData[:, 3]

        # lake simulation
        Lake.Qlake, _ = hbv_lake.simulate(
            plake,
            t,
            et,
            Lake.Parameters,
            [Model.conversionfactor, Lake.CatArea, Lake.LakeArea],
            Lake.StageDischargeCurve,
            0,
            init_st=Lake.InitialCond,
            ll_temp=tm,
            lake_sim=True,
        )
        # qlake is in m3/sec
        # lake routing
        Lake.QlakeR = routing.Muskingum_V(
            Lake.Qlake,
            Lake.Qlake[0],
            Lake.Parameters[11],
            Lake.Parameters[12],
            Model.conversionfactor,
        )

        # subcatchment
        distrrm.RunLumpedRRM(Model)

        # routing lake discharge with DS cell k & x and adding to cell Q
        qlake = routing.Muskingum_V(
            Lake.QlakeR,
            Lake.QlakeR[0],
            Model.Parameters[Lake.OutflowCell[0], Lake.OutflowCell[1], 10],
            Model.Parameters[Lake.OutflowCell[0], Lake.OutflowCell[1], 11],
            Model.conversionfactor,
        )

        qlake = np.append(qlake, qlake[-1])
        # both lake & Quz are in m3/s
        Model.quz[Lake.OutflowCell[0], Lake.OutflowCell[1], :] = (
            Model.quz[Lake.OutflowCell[0], Lake.OutflowCell[1], :] + qlake
        )

        # run the GIS part to rout from cell to another
        distrrm.SpatialRouting(Model)

        # Model.qout = Model.qout[:-1]

    @staticmethod
    def FW1(Model, ll_temp=None, q_0=None):
        """FW1.

        FW1 connects two module :
            1- The distributed rainfall-runoff module
            2- Triangular function-1 routing method
        Parameters
        ----------
        Model : TYPE
            DESCRIPTION.
        ll_temp : TYPE, optional
            DESCRIPTION. The default is None.
        q_0 : TYPE, optional
            DESCRIPTION. The default is None.

        Returns
        -------
        None.
        """

        # subcatchment
        distrrm.RunLumpedRRM(Model)

        distrrm.DistMaxbas1(Model)

        qlz1 = np.array(
            [np.nansum(Model.qlz[:, :, i]) for i in range(Model.TS)]
        )  # average of all cells (not routed mm/timestep)
        quz1 = np.array(
            [np.nansum(Model.quz[:, :, i]) for i in range(Model.TS)]
        )  # average of all cells (routed mm/timestep)

        Model.qout = qlz1 + quz1

        Model.qout = Model.qout[:-1]

    @staticmethod
    def FW1Withlake(Model, Lake, ll_temp=None, q_0=None):
        """FW1Withlake.

        FW1 connects two module :
            1- The distributed rainfall-runoff module
            2- Triangular function-1 routing method
            3- Lake module

        Parameters
        ----------
        Model : TYPE
            DESCRIPTION.
        Lake : TYPE
            DESCRIPTION.
        ll_temp : TYPE, optional
            DESCRIPTION. The default is None.
        q_0 : TYPE, optional
            DESCRIPTION. The default is None.

        Returns
        -------
        None.
        """

        plake = Lake.MeteoData[:, 0]
        et = Lake.MeteoData[:, 1]
        t = Lake.MeteoData[:, 2]
        tm = Lake.MeteoData[:, 3]

        # lake simulation
        Lake.Qlake, _ = hbv_lake.simulate(
            plake,
            t,
            et,
            Lake.Parameters,
            [Model.conversionfactor, Lake.CatArea, Lake.LakeArea],
            Lake.StageDischargeCurve,
            0,
            init_st=Lake.InitialCond,
            ll_temp=tm,
            lake_sim=True,
        )

        # qlake is in m3/sec
        # lake routing
        Lake.QlakeR = routing.muskingum(
            Lake.Qlake,
            Lake.Qlake[0],
            Lake.Parameters[11],
            Lake.Parameters[12],
            Model.conversionfactor,
        )

        # subcatchment
        distrrm.RunLumpedRRM(Model)

        distrrm.DistMAXBAS(Model)

        qlz1 = np.array(
            [
                np.nansum(Model.qlz[:, :, i])
                for i in range(Model.Parameters.shape[2] + 1)
            ]
        )  # average of all cells (not routed mm/timestep)
        quz1 = np.array(
            [
                np.nansum(Model.quz[:, :, i])
                for i in range(Model.Parameters.shape[2] + 1)
            ]
        )  # average of all cells (routed mm/timestep)

        qout = qlz1 + quz1

        # qout = (qlz1 + quz1) * Model.CatArea / (Model.conversionfactor* 3.6)

        Model.qout = qout[:-1] + Lake.QlakeR

    @staticmethod
    def Lumped(Model, Routing=0, RoutingFn=[]):
        """Lumped.

        Parameters
        ----------
        Model: [function]
            conceptual model and it should contain a function called simulate
            data: [numpy array]
                meteorological data as array with the first column as precipitation
                second as evapotranspiration, third as temperature and forth column as
                long term average temperature
            parameters: [numpy array]
                conceptual model parameters as array
            p2: [List]
                list of unoptimized parameters
                p2[0] = tfac, 1 for hourly, 0.25 for 15 min time step and 24 for daily time step
                p2[1] = catchment area in km2
            init_st: [list]
                initial state variables values [sp, sm, uz, lz, wc].
        Routing: [0 or 1]
            to decide wether t route the generated discharge hydrograph or not
        RoutingFn: [function]
            function to route the dischrge hydrograph.

        Returns
        -------
        st: [numpy array]
            3d array of the 5 state variable data for each cell
        q_lz: [numpy array]
            1d array of the calculated discharge.

        examples
        --------
            p2=[24, 1530]
            #[sp,sm,uz,lz,wc]
            init_st=[0,5,5,5,0]
            snow=0
        """
        ### input data validation
        assert callable(
            RoutingFn
        ), "routing function should be of type callable (function that takes arguments)"

        # data
        p = Model.data[:, 0]
        et = Model.data[:, 1]
        t = Model.data[:, 2]
        tm = Model.data[:, 3]

        # from the conceptual model calculate the upper and lower response mm/time step
        Model.quz, Model.qlz, Model.statevariables = Model.LumpedModel.Simulate(
            p,
            t,
            et,
            tm,
            Model.Parameters,
            init_st=Model.InitialCond,
            q_init=Model.q_init,
            snow=Model.Snow,
        )
        # q mm , area sq km  (1000**2)/1000/f/60/60 = 1/(3.6*f)
        # if daily tfac=24 if hourly tfac=1 if 15 min tfac=0.25
        Model.quz = Model.quz * Model.CatArea / Model.conversionfactor
        Model.qlz = Model.qlz * Model.CatArea / Model.conversionfactor

        Model.Qsim = Model.quz + Model.qlz

        if Routing != 0 and Model.Maxbas:
            Model.Qsim = RoutingFn(np.array(Model.Qsim[:-1]), Model.Parameters[-1])
        elif Routing != 0:
            Model.Qsim = RoutingFn(
                np.array(Model.Qsim[:-1]),
                Model.Qsim[0],
                Model.Parameters[-2],
                Model.Parameters[-1],
                Model.dt,
            )


if __name__ == "__main__":
    print("Wrapper")



import warnings

import numpy as np


class InstabilityWarning(UserWarning):
    """Issued when results may be unstable."""

    pass


warnings.simplefilter("always", InstabilityWarning)
warnings.simplefilter("always", UserWarning)


def SilenceNumpyWarning():
    np.warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning)


def SilenceShapelyWarning():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
